module SymbVal where

import List

data SymbFunc = 
  SymbFunc String Int --user-defined function, #parameters
  | AND --built-in functions
  | OR
  | XOR
  | INV 
  | ADD
  | SUB
  | MUL 
  | NEG  
  | ID
  | EQUALS      
  deriving (Eq)
instance Show SymbFunc
 where
  show (SymbFunc s n) = s
  show AND = "AND"
  show OR = "OR"
  show XOR = "XOR"
  show ID = "ID"
  show INV = "INV"
  show ADD = "ADD"
  show SUB = "SUB"
  show MUL = "MUL"
  show NEG = "NEG"
  show EQUALS = "EQUALS"

data SymbVal = 
  FuncApp SymbFunc [SymbVal] 
  | VBool Bool 
  | VInt Integer 
  | Var String 
  | VarIx String Int 
  | Unknown
 deriving (Eq)
instance Show SymbVal
 where
  show Unknown = ""
  show (FuncApp f fs) = (show f) ++ "(" ++ (concat (intersperse ", " (map show fs))) ++ ")"
  show (VBool True) = "1"
  show (VBool False) = "0"
  show (VInt n) = show n
  show (Var x) = x
  show (VarIx x n) = x ++ "_" ++ (show n)

showHTML :: SymbVal -> String
showHTML Unknown = ""
showHTML (FuncApp ADD fs) = "(" ++ (concat (intersperse "+" (map showHTML fs))) ++ ")"
showHTML (FuncApp MUL fs) = "(" ++ concatMap showHTML fs ++ ")"
showHTML (FuncApp s fs) = (show s) ++ "(" ++ (concat (intersperse ", " (map showHTML fs))) ++ ")"
showHTML (VBool True) = "1"
showHTML (VBool False) = "0"
showHTML (VInt n) = show n
showHTML (Var x) = x
showHTML (VarIx x n) = x ++ "<SUB>" ++ (show n) ++ "</SUB>"


open :: SymbVal
open = Unknown

nParams :: SymbFunc -> Int
nParams (ID) = 1
nParams (INV) = 1
nParams (NEG) = 1
nParams (SymbFunc s n) = n
nParams z = 2

combine :: SymbVal -> SymbVal -> SymbVal
combine Unknown y = y
combine x Unknown = x
combine x y = Unknown

combineM :: SymbVal -> SymbVal -> Maybe SymbVal
combineM Unknown y = Just y
combineM x Unknown = Just x
combineM x y = Nothing

sOR :: SymbVal -> SymbVal -> SymbVal
sOR (VInt n) y = Unknown
sOR x (VInt n) = Unknown
sOR (VBool p) (VBool q) = VBool (p || q) 
sOR (VBool True) y = VBool True
sOR x (VBool True) = VBool True
sOR (VBool False) y = y
sOR x (VBool False) = x
sOR (FuncApp OR xs) (FuncApp OR ys) = FuncApp OR (xs ++ ys)
sOR x (FuncApp OR vs) = FuncApp OR (x:vs)
sOR (FuncApp OR vs) y = FuncApp OR (y:vs)
sOR x y = FuncApp OR [x,y]

sAND :: SymbVal -> SymbVal -> SymbVal
sAND (VInt n) y = Unknown
sAND x (VInt n) = Unknown
sAND (VBool p) (VBool q) = VBool (p && q) 
sAND (VBool True) y = y
sAND x (VBool True) = x
sAND (VBool False) y = VBool False
sAND x (VBool False) = VBool False
sAND (FuncApp AND xs) (FuncApp AND ys) = FuncApp AND (xs ++ ys)
sAND x (FuncApp AND vs) = FuncApp AND (x:vs)
sAND (FuncApp AND vs) y = FuncApp AND (y:vs)
sAND x y = FuncApp AND [x,y]

sXOR :: SymbVal -> SymbVal -> SymbVal
sXOR (VInt n) y = Unknown
sXOR x (VInt n) = Unknown
sXOR (VBool True) (VBool False) = VBool True 
sXOR (VBool False) (VBool True) = VBool True 
sXOR (VBool True) (VBool True) = VBool False 
sXOR (VBool False) (VBool False) = VBool False 
sXOR (VBool True) y = sEQUALS y (VBool False)
sXOR x (VBool True) = sEQUALS x (VBool False)
sXOR (VBool False) y = sEQUALS y (VBool True)
sXOR x (VBool False) = sEQUALS x (VBool True)
sXOR (FuncApp XOR xs) (FuncApp XOR ys) = FuncApp XOR (xs ++ ys)
sXOR x (FuncApp XOR vs) = FuncApp XOR (x:vs)
sXOR (FuncApp XOR vs) y = FuncApp XOR (y:vs)
sXOR x y = FuncApp XOR [x,y]

sINV :: SymbVal -> SymbVal
sINV (VInt x) = Unknown
sINV (VBool x) = VBool (not x)
sINV (FuncApp INV vs) = FuncApp ID vs
sINV z = FuncApp INV [z]

sID :: SymbVal -> SymbVal
sID x = x

sEQUALS :: SymbVal -> SymbVal-> SymbVal
sEQUALS (VInt p) (VInt q) = VBool (p == q)
sEQUALS (VBool p) (VBool q) = VBool (p == q) 
sEQUALS (Var p) (Var q) = VBool (p == q)
sEQUALS (VarIx p n) (VarIx q m) = VBool ((p == q) && (n == m))
sEQUALS (FuncApp EQUALS xs) (FuncApp EQUALS ys) = FuncApp EQUALS (xs ++ ys)
sEQUALS x (FuncApp EQUALS vs) = FuncApp EQUALS (x:vs)
sEQUALS (FuncApp EQUALS vs) y = FuncApp EQUALS (y:vs)
sEQUALS x y = FuncApp EQUALS [x,y]

sADD :: SymbVal -> SymbVal -> SymbVal
sADD (VBool p) y = Unknown
sADD x (VBool q) = Unknown
sADD x (VInt 0) = x
sADD (VInt 0) y = y
sADD (VInt p) (VInt q) = VInt (p + q)
sADD (FuncApp ADD xs) (FuncApp ADD ys) = FuncApp ADD (xs ++ ys)
sADD x (FuncApp ADD vs) = FuncApp ADD (x:vs)
sADD (FuncApp ADD vs) y = FuncApp ADD (vs ++ [y])
sADD x y = FuncApp ADD [x,y]

sSUB :: SymbVal -> SymbVal -> SymbVal
sSUB (VBool p) y = Unknown
sSUB x (VBool q) = Unknown
sSUB x (VInt 0) = x
sSUB (VInt 0) y = sNEG y
sSUB (VInt p) (VInt q) = VInt (p - q)
sSUB (Var x) (Var y) | x == y = VInt 0
sSUB (VarIx x n) (VarIx y m) | ((x == y) && (m == n)) = VInt 0
sSUB x y = FuncApp SUB [x,y]

sMUL :: SymbVal -> SymbVal -> SymbVal
sMUL (VBool p) y = Unknown
sMUL x (VBool q) = Unknown
sMUL x (VInt 0) = VInt 0
sMUL (VInt 0) y = VInt 0
sMUL x (VInt 1) = x
sMUL (VInt 1) y = y
sMUL (VInt p) (VInt q) = VInt (p * q)
sMUL (FuncApp MUL xs) (FuncApp MUL ys) = FuncApp MUL (xs ++ ys)
sMUL x (FuncApp MUL vs) = FuncApp MUL (x:vs)
sMUL (FuncApp MUL vs) y = FuncApp MUL (vs ++ [y])
sMUL x y = FuncApp MUL [x,y]

sNEG :: SymbVal -> SymbVal
sNEG (VBool x) = Unknown
sNEG (VInt x) = VInt (negate x)
sNEG (FuncApp NEG vs) = FuncApp ID vs
sNEG z = FuncApp NEG [z]

lift1 :: (SymbVal -> SymbVal) -> SymbVal -> SymbVal
lift1 f Unknown = Unknown
lift1 f x = f x

lift2 :: (SymbVal -> SymbVal -> SymbVal) -> SymbVal -> SymbVal -> SymbVal
lift2 f Unknown y = Unknown
lift2 f x Unknown = Unknown
lift2 f x y = f x y


flatten :: SymbVal -> SymbVal
flatten (FuncApp f vs) = symbApply f (map flatten vs)
flatten z = z

symbApply :: SymbFunc -> [SymbVal] -> SymbVal
symbApply f vals = 
  if elem Unknown vals 
    then Unknown 
    else if elem open vals
      then open
      else app f vals
 where
  app OR vals = foldl (lift2 sOR) (VBool False) vals
  app AND vals = foldl (lift2 sAND) (VBool True) vals
  app XOR (v:vals) = foldl (lift2 sXOR) v vals
  app INV vals = (lift1 sINV) (head vals)
  app ID vals = (lift1 sID) (head vals)
  app EQUALS (v:vals) = foldl (lift2 sEQUALS) v vals
  app ADD vals = foldl (lift2 sADD) (VInt 0) vals
  app SUB (v:vals) = foldl (lift2 sSUB) v vals
  app MUL vals = foldl (lift2 sMUL) (VInt 1) vals
  app NEG vals = (lift1 sNEG) (head vals)
  app f vals = FuncApp f vals

equalNoIx :: SymbVal -> SymbVal -> Bool
equalNoIx (VarIx x n) (VarIx y m) = x == y
equalNoIx x y = x == y

getIxs :: SymbVal -> [(String, Int)]
getIxs (VarIx x n) = [(x,n)]
getIxs (FuncApp f vs) = concat (mapM getIxs vs)
getIxs z = []

groupIxs :: [(String, Int)] -> [(String, [Int])]
groupIxs [] = []
groupIxs ((s,n):ns) = (s, nub (n : (map snd (filter (fstEq s) ns)))) : (groupIxs ns')
 where
  ns' = filter (fstNeq s) ns
  fstEq x (y,s) = x == y
  fstNeq x (y,s) = x /= y
